"""
    Build script for lattice ice40 FPGAs
    latticeice40-builder.py
"""
import os
from os.path import join
from SCons.Script import (AlwaysBuild, Builder, DefaultEnvironment, Exit,
                          Environment, Default, Glob, COMMAND_LINE_TARGETS,
                          GetOption)

env = DefaultEnvironment()
env.Replace(PROGNAME="hardware")
env.Append(SIMULNAME="simulation")

# -- Get the local folder in which the icestorm tools should be installed
piopackages_dir = env.subst('$PIOPACKAGES_DIR')
bin_dir = join(piopackages_dir, 'toolchain-icestorm', 'bin')

# -- Add this path to the PATH env variable. First the building tools will be
# -- searched in the local PATH. If they are not founde, the global ones will
# -- be executed (if installed)
env.PrependENVPath('PATH', bin_dir)

# -- Target name for synthesis
TARGET = join(env['BUILD_DIR'], env['PROGNAME'])

# -- Target name for simulation
# TARGET_SIM = join(env['PROJECT_DIR'], env['SIMULNAME'])

# -- Get a list of all the verilog files in the src folfer, in ASCII, with
# -- the full path. All these files are used for the simulation
v_nodes = Glob(join(env['PROJECTSRC_DIR'], '*.v'))
src_sim = ["{}".format(f) for f in v_nodes]

# --------- Get the Testbench file (there should be only 1)
# -- Create a list with all the files finished in _tb.v. It should contain
# -- the test bench
list_tb = [f for f in src_sim if f[-5:].upper() == "_TB.V"]

if len(list_tb) > 1:
    print "---> WARNING: More than one testbenches used"

# -- Error checking
try:
    testbench = list_tb[0]

# -- there is no testbench
except IndexError:
    testbench = None

if 'sim' in COMMAND_LINE_TARGETS:
    if testbench is None:
        print "ERROR!!! NO testbench found for simulation"
        Exit(1)

    # -- Simulation name
    testbench_file = os.path.split(testbench)[-1]
    SIMULNAME, ext = os.path.splitext(testbench_file)
else:
    SIMULNAME = ''


TARGET_SIM = join(env.subst('$BUILD_DIR'), SIMULNAME)

# -------- Get the synthesis files.  They are ALL the files except the
# -------- testbench
src_synth = [f for f in src_sim if f not in list_tb]

# -- For debugging
print "Testbench: {}".format(testbench)

# -- Get the PCF file
src_dir = env.subst('$PROJECTSRC_DIR')
PCFs = join(src_dir, '*.pcf')
PCF_list = Glob(PCFs)

try:
    PCF = PCF_list[0]
except IndexError:
    print "\n--------> ERROR: no .pcf file found <----------\n"
    Exit(2)

# -- Debug
print "----> PCF Found: {}".format(PCF)

# -- Builder 1 (.v --> .blif)
synth = Builder(action='yosys -p \"synth_ice40 -blif {}.blif\" \
                        $SOURCES'.format(TARGET),
                suffix='.blif',
                src_suffix='.v')

# -- Builder 2 (.blif --> .asc)
pnr = Builder(action='arachne-pnr -d 1k -o $TARGET -p {} \
                      $SOURCE'.format(PCF),
              suffix='.asc',
              src_suffix='.blif')

# -- Builder 3 (.asc --> .bin)
bitstream = Builder(action='icepack $SOURCE $TARGET',
                    suffix='.bin',
                    src_suffix='.asc')

# -- Builder 4 (.asc --> .rpt)
time_rpt = Builder(action='icetime -mtr $TARGET $SOURCE',
                   suffix='.rpt',
                   src_suffix='.asc')

env.Append(BUILDERS={'Synth': synth, 'PnR': pnr, 'Bin': bitstream,
                     'Time': time_rpt})

blif = env.Synth(TARGET, [src_synth])
asc = env.PnR(TARGET, [blif, PCF])
binf = env.Bin(TARGET, asc)

upload = env.Alias('upload', binf, 'iceprog ' + ' $SOURCE')
AlwaysBuild(upload)

# -- Target for calculating the time (.rpt)
# rpt = env.Time(asc)
t = env.Alias('time', env.Time('time.rpt', asc))

# -------------------- Simulation ------------------
# -- Constructor para generar simulacion: icarus Verilog
iverilog = Builder(action='iverilog -o $TARGET $SOURCES ',
                   suffix='.out',
                   src_suffix='.v')

vcd = Builder(action=' $SOURCE',
              suffix='.vcd', src_suffix='.out')

simenv = Environment(BUILDERS={'IVerilog': iverilog, 'VCD': vcd},
                     ENV=os.environ)

out = simenv.IVerilog(TARGET_SIM, src_sim)
vcd_file = simenv.VCD(SIMULNAME, out)

waves = simenv.Alias('sim', vcd_file, 'gtkwave ' +
                     join(env['PROJECT_DIR'], "{} ".format(vcd_file[0])) +
                     join(env['PROJECTSRC_DIR'], SIMULNAME) +
                     '.gtkw')
AlwaysBuild(waves)

Default([binf])

# -- These is for cleaning the files generated using the alias targets
if GetOption('clean'):
    env.Default([t])
    simenv.Default([out, vcd_file])



import os
import shutil
import glob

from os.path import join, isdir, dirname

from .. import util


class PiofpgaInstaller(object):
    """Support for FPGA in platformio(pio) plug-in installer"""

    def install(self):
        print("Installing FPGA support for platformio...")
        pio_dir = os.path.join(dirname(__file__), 'platformio')

        # -- Source dirs
        board_dir = os.path.join(pio_dir, 'boards')
        builder_dir = os.path.join(pio_dir, 'builder', 'scripts')
        platform_dir = os.path.join(pio_dir, 'platforms')

        # -- Dest dirs
        dest_dir = util.get_home_dir()
        board_dest_dir = join(dest_dir, 'boards')
        platform_dest_dir = join(dest_dir, 'platforms')

        # -- Install board files
        self._copy_files(src=board_dir, dest=board_dest_dir)

        # -- Install platform files
        self._copy_files(src=platform_dir, dest=platform_dest_dir)

        # -- Install build script
        builder_files = glob.glob(join(builder_dir, '*'))
        for f in builder_files:
            path, name = os.path.split(f)
            name, ext = os.path.splitext(name)
            if not self._is_pyc(f):
                shutil.copy(f, join(platform_dest_dir, name + '-builder.py'))

    def _copy_files(self, src, dest):
        """Copy files from src to dest folder. Files .pyc are not copied"""

        # -- Check for the dest dir
        if isdir(dest):

            # -- It exists
            board_files = glob.glob(join(src, '*'))
            for f in board_files:
                if not self._is_pyc(f):
                    shutil.copy(f, dest)
                else:
                    print("Ignorig {}".format(f))
        else:
            # -- dest directory does not exist
            shutil.copytree(src, dest)

    def _is_pyc(self, filename):
        """return True if it is a .pyc file"""

        name, ext = os.path.splitext(filename)
        return ext.upper() == ".PYC"



# Profile class

import json
from os import makedirs
from os.path import isdir, isfile, join, expanduser


class Profile(object):

    def __init__(self):
        self.packages = {}
        self._apio_dir = join(expanduser("~"), '.apio/')
        self._profile_path = join(self._apio_dir, 'profile.json')
        self.load()

    def check_version(self, name, filename):
        return not ((name in self.packages.keys()) and (self.packages[name] >= filename))

    def remove(self, name):
        if name in self.packages.keys():
            del self.packages[name]

    def load(self):
        if isfile(self._profile_path):
            with open(self._profile_path, 'r') as profile:
                try:
                    self.packages = json.load(profile)
                except:
                    self.packages = {}
                profile.close()

    def save(self):
        if not isdir(self._apio_dir):
            makedirs(self._apio_dir)
        with open(self._profile_path, 'w') as profile:
            json.dump(self.packages, profile)
            profile.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import click

from .util import get_systype
from .examples import Examples
from .execute import SCons, System
from .packages.scons import SconsInstaller
from .packages.icestorm import IcestormInstaller
from .packages.driver import DriverInstaller
from .packages.system import SystemInstaller
from .packages.piofpga import PiofpgaInstaller

try:
    input = raw_input
except NameError:
    pass

try:
    unicode = str
except NameError:
    pass


@click.group()
@click.version_option()
def cli():
    """
    Environment for icestorm toolchain management
    """


@cli.command('debug')
def debug():
    """Show system information."""
    print('Platform: ' + get_systype())


@cli.command('init')
def init():
    """Create default SConstruct file."""
    SCons().create_sconstruct()

EXAMPLE_OF_USE_CAD = """
Example of use:
  apio examples -f leds
Copy the leds example files to the current directory
"""


@cli.command('examples')
@click.pass_context
@click.option('-l', '--list', is_flag=True, help='List all available examples.')
@click.option('-d', '--dir', type=unicode, metavar='NAME',
              help='Copy the selected example directory.')
@click.option('-f', '--files', type=unicode, metavar='NAME',
              help='Copy the selected example files.')
def examples(ctx, list, dir, files):
    """Manage default verilog examples."""
    if list:
        examples = Examples().list_examples()
        print('')
        for example in examples:
            print(' > ' + example)
        print('')
        print('To get and example, use the command:')
        print('apio examples -d/-f name')
        print(EXAMPLE_OF_USE_CAD)
    elif dir:
        Examples().copy_example_dir(dir)
    elif files:
        Examples().copy_example_files(files)
    else:
        print(ctx.get_help())
        print(EXAMPLE_OF_USE_CAD)


# System #


@cli.group()
def system():
    """System development tools.\n
       Install with `apio install system`"""


@system.command('lsusb')
def lsusb():
    """List all connected USB devices."""
    System().lsusb()


@system.command('lsftdi')
def lsftdi():
    """List all connected FTDI devices."""
    System().lsftdi()


# Install #


@cli.group('install', invoke_without_command=True)
@click.pass_context
@click.option('--all', is_flag=True, help='Install all toolchains')
def install(ctx, all):
    """Install development tools."""
    if ctx.invoked_subcommand is None:
        SystemInstaller().install()
        SconsInstaller().install()
        IcestormInstaller().install()


@install.command('driver')
def install_driver():
    """Install open FPGA drivers."""
    DriverInstaller().install()


@install.command('system')
def install_system():
    """Install system development tools."""
    SystemInstaller().install()


@install.command('scons')
def install_scons():
    """Install scons toolchain."""
    SconsInstaller().install()


@install.command('icestorm')
def install_icestorm():
    """Install icestorm toolchain."""
    IcestormInstaller().install()


@install.command('pio-fpga')
def intall_pio_fpga():
    """Install platformio-fpga support."""
    PiofpgaInstaller().install()
    print("> Now execute the following command:")
    print("")
    print("pio platforms install lattice_ice40")


# Uninstall #


@cli.group('uninstall', invoke_without_command=True)
@click.pass_context
@click.option('--all', is_flag=True, help='Uninstall all toolchains')
def uninstall(ctx, all):
    """Uninstall development tools."""
    if ctx.invoked_subcommand is None:
        _uninstall(SystemInstaller().uninstall,
                   SconsInstaller().uninstall,
                   IcestormInstaller().uninstall)


@uninstall.command('driver')
def uninstall_driver():
    """Uninstall open FPGA drivers."""
    _uninstall(DriverInstaller().uninstall)


@uninstall.command('system')
def uninstall_system():
    """Uninstall system development tools."""
    _uninstall(SystemInstaller().uninstall)


@uninstall.command('scons')
def uninstall_scons():
    """Uninstall scons toolchain."""
    _uninstall(SconsInstaller().uninstall)


@uninstall.command('icestorm')
def uninstall_icestorm():
    """Uninstall icestorm toolchain."""
    _uninstall(IcestormInstaller().uninstall)


def _uninstall(*functions):
    key = input('Are you sure? [Y/N]: ')
    if key == 'y' or key == 'Y':
        for count, function in enumerate(functions):
            function()


# Synthesize #


@cli.command('clean')
def clean():
    """Remove previous bitstream."""
    SCons().run(['-c'])


@cli.command('build')
def build():
    """Synthesize the bitstream."""
    SCons().run()


@cli.command('upload')
def upload():
    """Upload bitstream to FPGA."""
    SCons().run(['upload'])


@cli.command('time')
def time():
    """Bitstream timing analysis."""
    SCons().run(['time'])


@cli.command('sim')
def sim():
    """Launch verilog simulation."""
    SCons().run(['sim'])


if __name__ == '__main__':
    cli()



# Execute functions

import os
import glob
import click
import shutil

from os.path import join, isdir, isfile, dirname, basename

try:
    input = raw_input
except NameError:
    pass

# -- Error messages
EXAMPLE_NOT_FOUND_MSG = """
Sorry, this example does not exist
Use "apio examples -l" for listing all the available examples"""


class Examples(object):

    def __init__(self):
        self.examples_dir = join(dirname(__file__), '..', 'examples')

    def list_examples(self):
        return sorted(os.listdir(self.examples_dir))

    def copy_example_dir(self, example):
        example_path = join(os.getcwd(), example)
        local_example_path = join(self.examples_dir, example)

        if isdir(local_example_path):
            if isdir(example_path):
                click.echo('Warning: ' + example + ' directory already exists')
                key = input('Do you want to replace it? [Y/N]: ')
                if key == 'y' or key == 'Y':
                    shutil.rmtree(example_path)
                    self._copy_dir(example, local_example_path, example_path)
            else:
                self._copy_dir(example, local_example_path, example_path)
        else:
            click.echo(EXAMPLE_NOT_FOUND_MSG)

    def copy_example_files(self, example):
        example_path = os.getcwd()
        local_example_path = join(self.examples_dir, example)

        if isdir(local_example_path):
            self._copy_files(example, local_example_path, example_path)
        else:
            click.echo(EXAMPLE_NOT_FOUND_MSG)

    def _copy_files(self, example, src_path, dest_path):
        click.echo(' Copying ' + example + ' example')
        example_files = glob.glob(join(src_path, '*'))
        for f in example_files:
            filename = basename(f)
            if isfile(join(dest_path, filename)):
                click.echo('Warning: ' + filename + ' file already exists')
                key = input('Do you want to replace it? [Y/N]: ')
                if key == 'y' or key == 'Y':
                    shutil.copy(f, dest_path)
            else:
                shutil.copy(f, dest_path)

    def _copy_dir(self, example, src_path, dest_path):
        click.echo(' Creating ' + example + ' directory')
        shutil.copytree(src_path, dest_path)



import os
from platformio.platforms.base import BasePlatform


class Lattice_ice40Platform(BasePlatform):
    """
    The iCE40 family of ultra-low power, non-volatile FPGAs has five devices
    with densities ranging from 384 to 7680 Look-Up Tables (LUTs). In addition
    to LUT-based,low-cost programmable logic, these devices feature Embedded
    Block RAM (EBR), Non-volatile Configuration Memory (NVCM) and Phase Locked
    Loops (PLLs). These features allow the devices to be used in low-cost,
    high-volume consumer and system applications.

    http://www.latticesemi.com/Products/FPGAandCPLD/iCE40.aspx
    """

    PACKAGES = {

        "toolchain-icestorm": {

            # alias is used for quick access to package.
            "alias": "toolchain",

            # Flag which allows PlatformIO to install this package by
            # default via `> platformio install test` command
            "default": True
        },
    }

    def get_build_script(self):
        """ Returns a path to build script """

        # You can return static path
        # return "/path/to/test-builder.py"
        # or detect dynamically if `test-builder.py` is located in the same
        # folder with `test.py`
        return os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "lattice_ice40-builder.py")
